<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2012 Franz Holzinger <franz@ttproducts.de>
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/


/**
 * Debug extension.
 *
 * @author	Franz Holzinger <franz@ttproducts.de>
 * $Id: class.tx_fhdebug.php 186 2013-05-27 14:04:49Z franzholz $
 */
class tx_fhdebug {
	static public $prefixFieldArray =
		array(
			'file' => '',
			'line' => '#',
			'function' => '->'
		);
	static public $csConvObj;
	static public $errorLogFile = 'phpDebugErrorLog.txt';

	static protected $bActive = FALSE;	// inactive without initialization
	static protected $bInitialization = FALSE;
	static protected $bErrorWritten = FALSE;
	static protected $bUseErrorLog = FALSE;

	static private $phpVersionGt50205;
	static private $username;
	static private $bUserAllowed = TRUE;
	static private $extConf = array();
	static private $hndFile = 0;
	static private $bHasBeenInitialized = FALSE;
	static private $bNeedsFileInit = TRUE;
	static private $starttimeArray = array();
	static private $bCreateFile = FALSE;
	static private $hndProcessfile = FALSE;
	static private $processCount = 0;
	static private $bHtml = FALSE;
	static private $bWriteHeader = FALSE;
	static private $instanceCount = 0;
	static private $errorLogFilename = '';
	static private $debugFilename = '';


	public function __construct ($newExtConf) {
		self::$extConf = $newExtConf;
		self::$instanceCount++;
		self::$csConvObj = t3lib_div::makeInstance('t3lib_cs');
		if ($newExtConf['ERROR_LOG'] != '') {
			self::$errorLogFile = $newExtConf['ERROR_LOG'];
			self::$bUseErrorLog = TRUE;
		} else if ($newExtConf['DEBUGFILE'] != '') {
			self::$debugFilename = $newExtConf['DEBUGFILE'];
		}

		self::$errorLogFilename = t3lib_div::resolveBackPath(PATH_typo3conf . '../' . self::$errorLogFile);

//  error_log('tx_fhdebug::__construct : ' .  self::$debugFilename . chr(13), 3, self::getErrorLogFilename());

//  error_log('$newExtConf = '. print_r($newExtConf, TRUE) . chr(13),  3, self::getErrorLogFilename());
//  error_log('tx_fhdebug::__construct : ' .  print_r(tx_fhdebug::getTraceArray(4), TRUE) . chr(13), 3, self::getErrorLogFilename());
	}

	static public function debugControl (array $parameters) {
	}

	static public function getErrorLogFilename () {
		return self::$errorLogFilename;
	}

	static public function getDebugFilename () {
		return self::$debugFilename;
	}

	static public function hasError () {
		$result = (self::$bErrorWritten);
// error_log('hasError $result: ' . $result . chr(13), 3, self::getErrorLogFilename());

		return $result;
	}

	static public function writeHeader ($cssFilename) {
		$out = '
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
   "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
  <title>Debug generated by fh_debug</title>
  <meta http-equiv="content-type" content="text/html;charset=utf-8" />
  <link rel="stylesheet" href="../' . t3lib_extMgm::siteRelPath(FH_DEBUG_EXT) . 'res/' . $cssFilename . '" />
</head>

<body>
';

// error_log('writeHeader $cssFilename: ' . $cssFilename . chr(13), 3, self::getErrorLogFilename());

		$extConf = self::getExtConf();
		self::write($out, ($extConf['DEBUGFILE'] == ''));
	}

	static public function writeBodyEnd () {
		$out =
'</body>';

// error_log('writeBodyEnd ' . chr(13), 3, self::getErrorLogFilename());
		$extConf = self::getExtConf();
		self::write($out, ($extConf['DEBUGFILE'] == ''));
	}

	static public function readIpAddress () {
		$ipAddress = '';

// error_log ('readIpAddress $_SERVER ' . print_r($_SERVER, TRUE) . chr(13), 3, self::getErrorLogFilename());

		if (!empty($_SERVER['HTTP_CLIENT_IP'])) {
			$ipAddress = $_SERVER['HTTP_CLIENT_IP'];
		} else if (!empty($_SERVER['HTTP_X_FORWARDED_FOR'])) {
			$ips = explode(',', $_SERVER['HTTP_X_FORWARDED_FOR']);
			$ipAddress = trim($ips[count($ips) - 1]);
		} else {
			$ipAddress = t3lib_div::getIndpEnv('REMOTE_ADDR');
		}

// error_log ('readIpAddress ENDE $ipAddress ' . $ipAddress . chr(13), 3, self::getErrorLogFilename());
		return $ipAddress;
	}

	static public function verifyIpAddress (
		$ipAddress,
		$extConf
	) {
// error_log ('verifyIpAddress $ipAddress ' . $ipAddress . chr(13), 3, self::getErrorLogFilename());
// error_log ('verifyIpAddress $extConf ' . print_r($extConf, TRUE) . chr(13), 3, self::getErrorLogFilename());

		$result =
			(
				t3lib_div::cmpIP(
					$ipAddress,
					$extConf['IPADDRESS']
				)
			);

// error_log ('verifyIpAddress $result ' . $result . chr(13), 3, self::getErrorLogFilename());
		return $result;
	}

	static public function verifyFeusername (
		$username,
		$extConf
	) {
// error_log ('verifyFeusername $username ' . $username . chr(13), 3, self::getErrorLogFilename());
		$result = TRUE;

		if (
			TYPO3_MODE == 'FE' &&
			$extConf['FEUSERNAMES'] != ''
		) {
			$tmpArray = t3lib_div::trimExplode(',', $extConf['FEUSERNAMES']);
// error_log ('verifyFeusername $tmpArray ' . print_r($tmpArray, TRUE) . chr(13), 3, self::getErrorLogFilename());

			if (
				isset($tmpArray) &&
				is_array($tmpArray) &&
				in_array($username, $tmpArray) === FALSE
			) {
				$result = FALSE;
// error_log ('verifyFeusername $username not found. ' . chr(13), 3, self::getErrorLogFilename());
			}
		}

// error_log ('verifyFeusername $result ' . $result . chr(13), 3, self::getErrorLogFilename());
		return $result;
	}

	static public function verifyDebugMode (
		$extConf
	) {
		$dbgMode = ($extConf['TYPO3_MODE'] ? strtoupper($extConf['TYPO3_MODE']) : 'OFF');

		$bIsAllowed =
			(
				TYPO3_MODE == $dbgMode ||
				$dbgMode == 'ALL'
			);

		return $bIsAllowed;
	}

	static public function init ($ipAddress) {
// error_log ('init START ========================================== ' . chr(13), 3, self::getErrorLogFilename());

		if (self::hasBeenInitialized()) {
// error_log ('init Abbruch $bHasBeenInitialized', 3, self::getErrorLogFilename());
			return FALSE;
		}

		$phpVersion = phpversion();
		self::$phpVersionGt50205 = version_compare($phpVersion, '5.2.5', '>=');

		$extConf = self::getExtConf();
// error_log ('$extConf: ' . print_r($extConf, TRUE), 3, self::getErrorLogFilename());
		$backtrace = self::getTraceArray();

		if ($extConf['STARTFILES'] != '') {
			$startFileArray = t3lib_div::trimExplode(',', $extConf['STARTFILES']);
			$bStartFileFound = FALSE;
			if (is_array($startFileArray)) {
				foreach ($startFileArray as $startFile) {
					if ($backtrace['0']['file'] == $startFile) {
						$bStartFileFound = TRUE;
						break;
					}
				}
			}

			if (!$bStartFileFound) {
//  error_log('init backtrace: ' . print_r($backtrace, TRUE) . chr(13), 3, self::getErrorLogFilename());
//  error_log('init $extConf: ' . print_r($extConf, TRUE) . chr(13), 3, self::getErrorLogFilename());
//  error_log('init cancelled because no STARTFILES for "' . $backtrace['0']['file'] . '"' . chr(13), 3, self::getErrorLogFilename());
				return FALSE;
			}
		}

		$bResetFileFound = FALSE;
		if ($backtrace['0']['file'] == 'mod.php') {
			$bResetFileFound = TRUE;
		}
//  error_log('init $bResetFileFound: ' . $bResetFileFound . chr(13), 3, self::getErrorLogFilename());

		self::setIsInitialization(TRUE);

		if (!$extConf['DEBUGBEGIN']) {
			self::setActive(TRUE);
		}

		if (!self::$bUseErrorLog) {

			if ($extConf['HTML'] == '1') {
				self::$bHtml = TRUE;
			}
		}
//  error_log('init $ipAddress: ' . $ipAddress . chr(13), 3, self::getErrorLogFilename());

		if (t3lib_div::cmpIP($ipAddress, '127.0.0.1')) {
			if (
				t3lib_div::cmpIP(
					$ipAddress,
					$extConf['IPADDRESS']
				)
			) {
//  error_log('init IP matches' . self::$bHtml . chr(13), 3, self::getErrorLogFilename());
				// nothing
			} else {
//  error_log('init IP does not match' . self::$bHtml . chr(13), 3, self::getErrorLogFilename());
				self::$starttimeArray = array('no debugging possible', 'Attention: The server variable REMOTE_ADDR is set to local.');
				self::setActive(FALSE);
			}
		}

		self::setHasBeenInitialized(TRUE);
		self::setIsInitialization(FALSE);
// 		error_log ('init ENDE ========================================== '. chr(13), 3, self::getErrorLogFilename());
		return TRUE;
	}

	static public function initFile () {
// error_log ('initFile START ============= ' . chr(13), 3, self::getErrorLogFilename());

		$extConf = self::getExtConf();

		if (self::$bUserAllowed && self::getDebugFilename() != '') {

			$processFilename = self::getProcessFilename();
// 	 error_log ('initFile $processFilename = ' . $processFilename . chr(13), 3, self::getErrorLogFilename());

			$readBytes = 0;
			if (!is_writable($processFilename)) {
				self::$hndProcessfile = fopen($processFilename, 'w+b');
				$readBytes = 0;
			} else {
				self::$hndProcessfile = fopen($processFilename, 'r+b');
				$readBytes = filesize($processFilename);
			}
// error_log ('initFile $readBytes = ' . $readBytes);
			if (self::$hndProcessfile) {
				if ($readBytes) {
					$processCount = intval(fread(self::$hndProcessfile, $readBytes));
					$processCount++;
// 	error_log ('initFile $processCount = ' . $processCount . ' Pos 1 ' . chr(13), 3, self::getErrorLogFilename());
				} else {
					$processCount = 1;
// 	error_log ('initFile $processCount = ' . $processCount . ' Pos 2 ' .  chr(13), 3, self::getErrorLogFilename());
				}

				if (
					$bResetFileFound ||
					$processCount > intval($extConf['APPENDDEPTH'])
				) {
					$processCount = 1;
// 	error_log ('initFile $processCount = ' . $processCount . ' Pos 3 ' .  chr(13), 3, self::getErrorLogFilename());
					self::setCreateFile();
				}
// 	error_log ('initFile write $processCount = ' . $processCount  .  ' Pos 8 ' . chr(13), 3, self::getErrorLogFilename());
				self::writeTemporaryFile($processCount);
			}
			$extPath = PATH_typo3conf;
			$filename = t3lib_div::resolveBackPath($extPath . '../' . self::getDebugFilename());
//  	error_log ('initFile write $filename = ' . $filename . chr(13), 3, self::getErrorLogFilename());
			$path_parts = pathinfo($filename);

			if (is_writable($path_parts['dirname'])) {
				self::$bWriteHeader = self::$bHtml;
//  	error_log ('initFile write self::$bWriteHeader = ' . self::$bWriteHeader . chr(13), 3, self::getErrorLogFilename());

				if ($extConf['APPENDDEPTH'] > 1) {
					if (self::$bCreateFile) {
						$openMode = 'w+b';
// error_log('initFile $openMode Pos 1 = ' . $openMode . chr(13), 3, self::getErrorLogFilename());
					} else {
						$openMode = 'a+b';
// error_log('initFile $openMode Pos 2 = ' . $openMode . chr(13), 3, self::getErrorLogFilename());
					}
				} else {
					$openMode = $extConf['DEBUGFILEMODE'];
// error_log('initFile $openMode Pos 3 = ' . $openMode . chr(13), 3, self::getErrorLogFilename());
				}
// error_log ('initFile fopen(' . $filename . ', ' . $openMode . ') ' . chr(13), 3, self::getErrorLogFilename() );

				self::$hndFile = fopen($filename, $openMode);

				if (self::$hndFile !== FALSE) {

					$ipAddress = self::readIpAddress();
					self::$starttimeArray =
						array(
							date('H:i:s  d.m.Y') . '  (' . $ipAddress . ')',
							'start date, time and IP of debug session (mode "' . $openMode . '")'
						);
				} else {
// error_log('initFile no file handle ERROR = ' . $out . chr(13), 3, self::getErrorLogFilename());
				}
			} else {
				self::setActive(FALSE); // no debug is necessary when the file cannot be written anyways
			}

			if (
				$extConf['DEVLOG'] &&
				!self::$hndFile &&
				!is_writable($filename)
			) {
				t3lib_div::devLog(
					'DEBUGFILE: "' . $filename . '" is not writable in mode="' . $openMode . '"',
					FH_DEBUG_EXT,
					0
				);

				t3lib_div::sysLog(
					'DEBUGFILE: "' . $filename . '" is not writable in mode="' . $openMode . '"',
					FH_DEBUG_EXT,
					0
				);
			}
		}
	}

	static public function getProcessFilename () {
		$result = PATH_site . 'typo3temp/fh_debug.txt';
		return $result;
	}

	static public function bIsActive () {

// error_log ('bIsActive self::$bActive = ' .  self::$bActive . chr(13), 3, self::getErrorLogFilename());
		return self::$bActive;
	}

	static public function setActive ($v) {
// if ($v) {
// error_log ('function setActive = ' . $v  . chr(13), 3, self::getErrorLogFilename() );
//
// $backtrace = self::getTraceArray();
// error_log('setActive backtrace: ' . print_r($backtrace, TRUE) . chr(13), 3, self::getErrorLogFilename());
// }

		self::$bActive = $v;
	}

	static public function setIsInitialization ($bInitialization) {
		self::$bInitialization = $bInitialization;
	}

	static public function bIsInitialization () {
		return self::$bInitialization;
	}

	static public function setHasBeenInitialized ($bHasBeenInitialized) {
		self::$bHasBeenInitialized = $bHasBeenInitialized;
// error_log('setHasBeenInitialized self::$bHasBeenInitialized: ' . self::$bHasBeenInitialized . chr(13), 3, self::getErrorLogFilename());
	}

	static public function hasBeenInitialized () {
// error_log('hasBeenInitialized self::$bHasBeenInitialized: ' . self::$bHasBeenInitialized . chr(13), 3, self::getErrorLogFilename());

		return self::$bHasBeenInitialized;
	}

	static public function truncateFile () {

// 		if (self::$hndFile) {
// 			self::$hndFile = ftruncate(self::$hndFile, 0);
// 			self::writeTemporaryFile(0);
// 			self::setHasBeenInitialized(FALSE);
// 		}
	}

	static public function setCreateFile () {

		self::$bCreateFile = TRUE;
	}

	/* determines if the PHP function debug_backtrace() may be called with the parameter to not populate the object index */
	static public function hasBacktraceParam () {
		return self::$phpVersionGt50205;
	}

	static public function debugBegin () {
// error_log('debugBegin'. chr(13), 3, self::getErrorLogFilename());

		if (self::hasBeenInitialized() && !self::hasError()) {
			$extConf = self::getExtConf();

			if ($extConf['DEBUGBEGIN']) {
				self::setActive(TRUE);

				$ipAddress = self::readIpAddress();
				self::debug(
					'debugBegin (' . $ipAddress . ') BEGIN >----- ----- ----->', 'debugBegin',
					'',
					'',
					TRUE
				);
/*
$backtrace = self::getTraceArray();
error_log('debugBegin backtrace: ' . print_r($backtrace, TRUE) . chr(13), 3, self::getErrorLogFilename());
*/

			}
		}
	}

	static public function debugEnd () {
// error_log('debugEnd' . chr(13), 3, self::getErrorLogFilename());
		if (self::hasBeenInitialized() ) {
			$extConf = self::getExtConf();

			if ($extConf['DEBUGBEGIN']) {
				$ipAddress = self::readIpAddress();
				self::debug(
					'debugEnd (' . $ipAddress . ') END <----- ----- -----<', 'debugEnd',
					'',
					'',
					TRUE
				);
				self::setActive(FALSE);

// $backtrace = self::getTraceArray();
// error_log('debugEnd backtrace: ' . print_r($backtrace, TRUE) . chr(13), 3, self::getErrorLogFilename());
			}
		}
	}

	static public function getExtConf () {
		$rc = self::$extConf;

		return $rc;
	}

	static public function getTraceFieldArray () {
		$extConf = self::getExtConf();
		$rc = t3lib_div::trimExplode(',', $extConf['TRACEFIELDS']);

		return $rc;
	}

	static public function getTraceArray ($depth = 0, $offset = 2) {

		if (self::hasBacktraceParam()) {
			$trail = debug_backtrace(FALSE);
		} else {
			$trail = debug_backtrace();
		}

		$last = count($trail) - 1;

		if (!$depth) {
			$depth = $last + 1;
			$offset = 0;
		}

		$traceFieldArray = self::getTraceFieldArray();
		$traceArray = array();
		$j = $depth - 1;

		for ($i = $offset; $i <= $last ; ++$i) {
			if (!isset($trail[$i])) {
				continue;
			}
			$theTrail = $trail[$i];
			if (!is_array($theTrail)) {
				continue;
			}

			foreach ($traceFieldArray as $traceField) {
				$v =
					(
						$traceField == 'file' && $theTrail['file'] != '' ?
							basename($theTrail[$traceField]) :
							$theTrail[$traceField]
					);
				$traceArray[$j][$traceField] = $v;
			}
			$j--;

			if ($j < 0) {
				break;
			}
		}
		ksort($traceArray);

		return $traceArray;
	}

	static public function printTraceLine ($traceArray) {
		$rc = '';
		$debugTrail = array();
		$extConf = self::getExtConf();

		if (is_array($traceArray) && count($traceArray)) {
			foreach ($traceArray as $i => $trace) {
				if (self::$bHtml) {
					$debugTrail[$i] .= '<tr>';
					foreach ($trace as $field => $v) {
						$debugTrail[$i] .= '<td>'; //  bgcolor="#E79F9F"
						$debugTrail[$i] .=  self::$prefixFieldArray[$traceField] . $v;
						$debugTrail[$i] .= '</td>';
					}
					$debugTrail[$i] .= '</tr>';
				} else {
					$debugTrail[$i] .= '|';
					foreach ($trace as $field => $v) {
						$debugTrail[$i] .=  self::$prefixFieldArray[$traceField] . $v;
						$debugTrail[$i] .= '|';
					}
					$debugTrail[$i] .= chr(13);
				}
			}
			$rc = implode('', $debugTrail);
			if (self::$bHtml) {
				$rc = '<table>' . $rc . '</table>';
			} else {
				$rc = chr(13) . '==============================' . chr(13) . $rc . chr(13);
			}
		}
		return $rc;
	}

	static public function printArrayVariable ($variable, $depth) {

		$extConf = self::getExtConf();

		if ($depth < $extConf['LEVEL']) {

			$debugArray = array();
			if (self::$bHtml) {
				foreach ($variable as $k => $v1) {
					$debugArray[$k] .= '<tr>';
					$td = '<td>';
					$debugArray[$k] .= $td;
					$debugArray[$k] .=  nl2br(htmlspecialchars($k));
					$debugArray[$k] .= '</td>';
					if (is_array($v1)) {
						$debugArray[$k] .= '<td>';
						$debugArray[$k] .= self::printArrayVariable($v1, $depth + 1);
						$debugArray[$k] .= '</td>';
					} else if (is_object($v1)) {
						$debugArray[$k] .= '<td>';
						$debugArray[$k] .= self::printObjectVariable($v1, $depth + 1);
						$debugArray[$k] .= '</td>';
					} else {
						$td = '<td>';
						$debugArray[$k] .= $td . nl2br(htmlspecialchars($v1)) . '</td>';
					}
					$debugArray[$k] .= '</tr>';
				}
			} else {
				foreach ($variable as $k => $v1) {
					$debugArray[$k] .=  $k;
					$debugArray[$k] .= '|';
					if (is_array($v1)) {
						$debugArray[$k] .= self::printArrayVariable($v1, $depth + 1);
					} else if (is_object($v1)) {
						$debugArray[$k] .= self::printObjectVariable($v1, $depth + 1);
					} else {
						$debugArray[$k] .=  $v1;
					}
					$debugArray[$k] .= '|' . chr(13);
				}
			}

			$rc = implode('', $debugArray);

			if (self::$bHtml) {
				$rc = '<table>' . $rc . '</table>';
			}
		} else {
			$rc = '->...';
		}
// error_log ('printArrayVariable $rc = ' . $rc);
		return $rc;
	}

	static public function printObjectVariable ($variable, $depth) { // TODO: show private member variables

		$vars = array();
		$cObjArray = new ArrayObject($variable);
		$vars = (array) @get_object_vars($cObjArray);

		$rc = self::printArrayVariable($vars, $depth);

		return $rc;
	}

	static public function printVariable ($variable) {
		$rc = '';
		$debugArray = array();
		$extConf = self::getExtConf();

		if (is_array($variable)) {
			$rc = self::printArrayVariable($variable, 0);
		} else if (is_object($variable))	{
			$rc = self::printObjectVariable($variable, 0);
		} else {
			if (self::$bHtml) {
				$rc = '<p>' . nl2br(htmlspecialchars($variable)) . '</p>';
			} else {
				$rc = $variable;
			}
		}
		return $rc;
	}

	static public function processUser ($extConf) {
		if (
			TYPO3_MODE == 'FE' &&
			$extConf['FEUSERNAMES'] != '' &&
			isset($GLOBALS['TSFE']) &&
			is_object($GLOBALS['TSFE'])
		) {
			if (is_array($GLOBALS['TSFE']->fe_user->user)) {
				$username = $GLOBALS['TSFE']->fe_user->user['username'];
			}

			if ($username != self::$username) {
				$bAllowFeuser = self::verifyFeusername(
					$username,
					$extConf
				);

				self::$bUserAllowed = $bAllowFeuser;
				if ($bAllowFeuser) {
					self::$username = $username;
				}
			}
		}
	}

	static public function getTypeView ($variable) {
		$result = '';
		$type = gettype($variable);
		switch ($type) {
			case 'array':
				$result = ' (' . $type . ' of ' . count($variable) . ' items )';
			break;
			case 'object':
				$result = ' (' . $type . ' of class ' . get_class($variable) . ')';
			break;
			default:
				$result = ' (' . $type . ')';
		}
		return $result;
	}

	static public function writeTemporaryFile ($processCount) {
		$processFilename = self::getProcessFilename();
		if (!self::$hndProcessfile) {
			self::$hndProcessfile = fopen($processFilename, 'r+');
		}
		ftruncate(self::$hndProcessfile, 0);
		rewind(self::$hndProcessfile);
		fwrite(self::$hndProcessfile, $processCount);
		self::$processCount = $processCount;
		fclose(self::$hndProcessfile);
	}

	static public function write ($out, $bPrintOnScreen) {
// error_log('write START ' . chr(13), 3, self::getErrorLogFilename());
		$result = FALSE;

		if (self::$bUseErrorLog) {
// error_log('write self::$bUseErrorLog = TRUE ' . chr(13), 3, self::getErrorLogFilename());
			$result = TRUE;
			 // keep the following line
			error_log($out . chr(13), 3, self::getErrorLogFilename());
		} else {
			if (self::$hndFile) {
// error_log('write vor fputs ' . chr(13), 3, self::getErrorLogFilename());
				fputs(self::$hndFile, $out);
				$result = TRUE;
// error_log('write nach fputs $out = ' . $out . chr(13), 3, self::getErrorLogFilename());
			} else if ($bPrintOnScreen) {
// error_log('write kein DEBUGFILE ' . chr(13), 3, self::getErrorLogFilename());
				echo $out;
				$result = TRUE;
			} else {
// error_log('write no file handle ERROR = ' . $out . chr(13), 3, self::getErrorLogFilename());
			}
		}
// error_log('write ENDE $result = ' . $result . chr(13), 3, self::getErrorLogFilename());

		return $result;
	}

	static public function writeOut ($extConf, $variable, $name, $bTrace = 1, $bHeader = FALSE) {
// error_log('writeOut $extConf ' . print_r($extConf, TRUE) . chr(13), 3, self::getErrorLogFilename());

		$type = '';
		$out = '';

		if ($bHeader) {
			$type = self::getTypeView($variable);
		}

// error_log('writeOut $variable ' . print_r($variable, TRUE) . chr(13), 3, self::getErrorLogFilename());
// error_log('writeOut $name ' . $name . chr(13), 3, self::getErrorLogFilename());

		if (
			self::$hndFile ||
			self::$bUseErrorLog ||
			$extConf['DEBUGFILE'] == ''
		) {
			$traceArray = ($bTrace ? self::getTraceArray($extConf['TRACEDEPTH']) : array());
// error_log('writeOut $traceArray ' . print_r($traceArray, TRUE) . chr(13), 3, self::getErrorLogFilename());
			$content = self::printTraceLine($traceArray);

// error_log('writeOut $content ' . $content . chr(13), 3, self::getErrorLogFilename());

			if (self::$bHtml) {
				$out = $content . '<br/>' .
					self::printVariable($variable) . chr(13) .
					'<h3>' . $name . $type . '</h3>' .
					'<hr/>' . chr(13);
			} else {
				$out = $content . '|' .
					self::printVariable($variable) . chr(13) .
					'###' . $name . $type . '###' . chr(13) .
					'--------------------------------------------' . chr(13);
			}
// error_log('writeOut $out ' . $out . chr(13), 3, self::getErrorLogFilename());
		}

		if (
			function_exists("mb_check_encoding") &&
			is_callable("mb_check_encoding")
		) {
			$charset = mb_detect_encoding($out, 'UTF-8,ASCII,ISO-8859-1,ISO-8859-15', TRUE);
		}

		if (
			$charset != '' &&
			$charset != 'UTF-8' &&
			$GLOBALS['TYPO3_CONF_VARS']['SYS']['t3lib_cs_convMethod'] != ''
		) {
			$out = self::$csConvObj->conv($out, $charset, 'UTF-8');
		}

		$bWritten = self::write($out, ($extConf['DEBUGFILE'] == ''));
//  error_log('debug nach write $bWritten = ' . $bWritten . chr(13), 3, self::getErrorLogFilename());

		if (
			!$bWritten &&
			!self::$bErrorWritten
		) {
			$overwriteModeArray = array('x', 'x+', 'xb', 'x+b');
			if (
				file_exists($extConf['DEBUGFILE']) &&
				in_array($extConf['DEBUGFILEMODE'], $overwriteModeArray)
			) {
				echo '<b>DEBUGFILE: "' . $extConf['DEBUGFILE'] . '" is not empty.</b>';
			} else {
				echo '<b>DEBUGFILE: "' . $extConf['DEBUGFILE'] . '" is not writable.</b>';
			}
			self::$bErrorWritten = TRUE;
//  error_log('debug self::$bErrorWritten = ' . self::$bErrorWritten . chr(13), 3, self::getErrorLogFilename());
		}

		return $bWritten;
	}

	static public function debug (
		$variable = '',
		$name = '*variable*',
		$line = '*line*',
		$file = '*file*',
		$bTrace = 1,
		$debugLevel = E_DEBUG
	) {
// error_log('tx_fhdebug::debug ================ ' . chr(13), 3, self::getErrorLogFilename());

		$bControlMode = FALSE;
		$charset = '';
		$extConf = self::getExtConf();
		$bUsernameIsAllowed = FALSE;

		if ($name == 'control:resetTemporaryFile') {
			self::truncateFile();
			$bControlMode = TRUE;
		}
		$storeIsActive = self::bIsActive();


// $backtrace = self::getTraceArray();
// error_log('tx_fhdebug::debug backtrace: ' . print_r($backtrace, TRUE) . chr(13), 3, self::getErrorLogFilename());

// error_log('tx_fhdebug::debug $variable = ' . print_r($variable, TRUE) . chr(13), 3, self::getErrorLogFilename());
// error_log('tx_fhdebug::debug $name = ' . print_r($name, TRUE) . chr(13), 3, self::getErrorLogFilename());

// error_log('tx_fhdebug::debug self::$bUseErrorLog = ' . self::$bUseErrorLog . chr(13), 3, self::getErrorLogFilename());

// error_log('tx_fhdebug::debug $storeIsActive = ' . $storeIsActive . chr(13), 3, self::getErrorLogFilename());

		self::processUser($extConf);

		if (
			!$bControlMode &&
			($storeIsActive || self::bIsInitialization())
		) {
			self::setActive(FALSE);

//  error_log('debug self::$bUserAllowed: ' . self::$bUserAllowed . chr(13), 3, self::getErrorLogFilename());

			if (self::$bUserAllowed) {

				if (self::$bNeedsFileInit) {
					self::initFile();
					self::$bNeedsFileInit = FALSE;
				}

// error_log('debug self::$bWriteHeader = ' . self::$bWriteHeader . chr(13), 3, self::getErrorLogFilename());
				if (self::$bWriteHeader) {
// error_log('debug self::$processCount = ' . self::$processCount . chr(13), 3, self::getErrorLogFilename());

					$headerPostFix = '';
					$headerValue = '';

//  error_log('debug self::$bWriteHeader = ' . self::$bWriteHeader . chr(13), 3, self::getErrorLogFilename());

					self::writeHeader($extConf['CSSFILE']);
					self::$bWriteHeader = FALSE;
//  error_log('nach writeHeader self::$bWriteHeader = ' . self::$bWriteHeader . chr(13), 3, self::getErrorLogFilename());

// error_log('self::$starttimeArray: ' . print_r(self::$starttimeArray, TRUE) . chr(13), 3, self::getErrorLogFilename());
					if (count(self::$starttimeArray)) {
						$headerPostFix = self::$starttimeArray['1'];
						$headerValue = self::$starttimeArray['0'];
						self::$starttimeArray = array();
					}

					$appendText = ' - counter: ' . self::$processCount . ' ' . $headerPostFix;

					switch (TYPO3_MODE) {
						case 'FE':
							if (self::$bHtml) {
								$head = '<h3>Front End Debugging<br />' . $appendText . '</h3>';
							} else {
								$head = '#Front End Debugging' . $appendText . '#';
							}
							break;
						case 'BE':
							if (self::$bHtml) {
								$head = '<h3>Back End Debugging<br />' . $appendText . '</h3>';
							} else {
								$head = '#Back End Debugging' . $appendText . '#';
							}
							break;
					}
//  error_log('debug $headerValue = ' . print_r($headerValue, TRUE) . chr(13), 3, self::getErrorLogFilename());
//  error_log('debug $head = ' . $head . chr(13), 3, self::getErrorLogFilename());

					self::writeOut($extConf, $headerValue, $head, FALSE, TRUE);
				}
				self::writeOut($extConf, $variable, $name, $bTrace, FALSE);
			}
			self::setActive($storeIsActive);
		}
	}

	/**
	 * Returns the internal debug messages as a string.
	 *
	 * @return string
	 */
	static public function toString () {

		$filename = '';

		if (self::$bUseErrorLog) {
			$filename = self::getErrorLogFilename();
		}
		if ($filename == '') {
			$filename = self::getDebugFilename();
		}

		$result = 'The debug messages have been written to the file "' . $filename . '"';

		return $result;
	}


	public function __destruct () {
//  error_log('__destruct' . chr(13), 3, self::getErrorLogFilename());

		if (self::$hndFile) {
			self::$instanceCount--;
//  error_log('__destruct self::$instanceCount ' . self::$instanceCount . chr(13), 3, self::getErrorLogFilename());

			$extConf = self::getExtConf();
			if (!self::$instanceCount && $extConf['APPENDDEPTH'] == '0') {
				self::writeBodyEnd();
			}

			if (!self::$instanceCount) {
				fclose(self::$hndFile);
				self::setHasBeenInitialized(FALSE);
				self::$hndFile = NULL; // this is a static class which remains even after the closing of the object
//  error_log('__destruct delete $hndFile ' . chr(13), 3, self::getErrorLogFilename());
			} else {
				fflush(self::$hndFile);
			}
		}
//  error_log('__destruct ENDE' . chr(13), 3, self::getErrorLogFilename());

// 		ob_flush(); // flush the error_log
// 		flush();
	}
}


/**
 * This function provides a shortcut so you don't have to use the class directly
 */
function fhdebug (
	$variable = '',
	$name = '*variable*',
	$line = '*line*',
	$file = '*file*',
	$traceDepth = 3,
	$debugLevel = E_DEBUG
) {
	global $TYPO3_CONF_VARS, $myDebugObject;

	$myDebugObject->debug(
		$variable,
		$name,
		$line,
		$file,
		$traceDepth,
		$debugLevel
	);
}


if (defined("TYPO3_MODE") && $TYPO3_CONF_VARS[TYPO3_MODE]['XCLASS']['ext/fh_debug/lib/class.tx_fhdebug.php'])	{
	include_once($TYPO3_CONF_VARS[TYPO3_MODE]['XCLASS']['ext/fh_debug/lib/class.tx_fhdebug.php']);
}

?>